import sys
import os
import socket
import time
import select
import threading
from PyQt5.QtCore import QThread, pyqtSignal

# Add Protocol directory to Python path
sys.path.insert(0, os.path.join(os.path.dirname(os.path.dirname(__file__)), 'Protocol'))

from protocol import IcingProtocol
from voice_codec import VoiceProtocol, Codec2Mode
from messages import VoiceStart, VoiceAck, VoiceEnd
from encryption import EncryptedMessage

class ProtocolPhoneClient(QThread):
    """Phone client that integrates the full Icing Protocol with 4FSK and ChaCha20."""
    
    data_received = pyqtSignal(bytes, int)
    state_changed = pyqtSignal(str, str, int)
    audio_received = pyqtSignal(bytes, int)  # For decoded audio
    
    def __init__(self, client_id, identity_keys=None):
        super().__init__()
        self.client_id = client_id
        self.running = True
        
        # Initialize Icing Protocol
        self.protocol = IcingProtocol()
        
        # Override identity keys if provided
        if identity_keys:
            self.protocol.identity_privkey = identity_keys[0]
            self.protocol.identity_pubkey = identity_keys[1]
        
        # Connection state
        self.connected = False
        self.handshake_complete = False
        
        # Voice state
        self.voice_active = False
        self.voice_protocol = None
        
        # Peer information
        self.peer_identity_hex = None
        self.peer_port = None
        
        # For GSM simulator compatibility
        self.gsm_host = "localhost"
        self.gsm_port = 12345
        self.gsm_socket = None
        
        # Track processed messages
        self.processed_message_count = 0
        
    def set_peer_identity(self, peer_identity_hex):
        """Set the peer's identity public key (hex string)."""
        self.peer_identity_hex = peer_identity_hex
        self.protocol.set_peer_identity(peer_identity_hex)
        
    def set_peer_port(self, port):
        """Set the peer's port for direct connection."""
        self.peer_port = port
        
    def connect_to_gsm_simulator(self):
        """Connect to the GSM simulator for voice channel simulation."""
        try:
            self.gsm_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.gsm_socket.settimeout(5)
            self.gsm_socket.connect((self.gsm_host, self.gsm_port))
            print(f"Client {self.client_id} connected to GSM simulator")
            return True
        except Exception as e:
            print(f"Client {self.client_id} failed to connect to GSM simulator: {e}")
            return False
    
    def run(self):
        """Main thread loop."""
        # Protocol listener already started in __init__ of IcingProtocol
        
        # Connect to GSM simulator if available
        self.connect_to_gsm_simulator()
        
        while self.running:
            try:
                # Process protocol messages
                self._process_protocol_messages()
                
                # Process GSM simulator data if connected
                if self.gsm_socket:
                    self._process_gsm_data()
                    
                self.msleep(10)
                
            except Exception as e:
                print(f"Client {self.client_id} error in main loop: {e}")
                # Only emit state change if it's a real connection error
                if "get_messages" not in str(e):
                    self.state_changed.emit("CALL_END", "", self.client_id)
                break
                
    def _process_protocol_messages(self):
        """Process messages from the Icing Protocol."""
        # Process new messages in the inbound queue
        if not hasattr(self.protocol, 'inbound_messages'):
            return
            
        new_messages = self.protocol.inbound_messages[self.processed_message_count:]
        
        for msg in new_messages:
            self.processed_message_count += 1
            msg_type = msg.get('type', '')
            
            if msg_type == 'PING_REQUEST':
                # Received ping request, we're the responder
                if not self.protocol.state['ping_sent']:
                    # Enable auto responder to handle protocol flow
                    self.protocol.auto_responder = True
                    # Send ping response
                    index = self.protocol.inbound_messages.index(msg)
                    self.protocol.respond_to_ping(index, 1)  # Accept with ChaCha20
                    
            elif msg_type == 'PING_RESPONSE':
                # Ping response received, continue with handshake
                if not self.protocol.state['handshake_sent']:
                    self.protocol.send_handshake()
                    
            elif msg_type == 'HANDSHAKE':
                # Handshake received
                if self.protocol.state['ping_sent'] and not self.protocol.state['handshake_sent']:
                    # We're initiator, send our handshake
                    self.protocol.send_handshake()
                # Derive keys if we have peer's handshake
                if self.protocol.state['handshake_received'] and not self.protocol.state['key_exchange_complete']:
                    self.protocol.derive_hkdf()
                    self.handshake_complete = True
                    self.state_changed.emit("HANDSHAKE_DONE", "", self.client_id)
                    
            elif msg_type == 'ENCRYPTED':
                # Decrypt and process encrypted message
                parsed = msg.get('parsed')
                if parsed and hasattr(parsed, 'plaintext'):
                    self._handle_encrypted_message(parsed.plaintext)
                    
            elif msg_type == 'voice_start':
                # Voice session started by peer
                self._handle_voice_start(msg.get('parsed'))
                
            elif msg_type == 'voice_ack':
                # Voice session acknowledged
                self.voice_active = True
                
            elif msg_type == 'voice_end':
                # Voice session ended
                self.voice_active = False
                
    def _process_gsm_data(self):
        """Process audio data from GSM simulator."""
        try:
            readable, _, _ = select.select([self.gsm_socket], [], [], 0)
            if readable:
                data = self.gsm_socket.recv(4096)
                if data and self.voice_active and self.voice_protocol:
                    # Process received FSK-modulated audio
                    encrypted_frames = self.voice_protocol.demodulate_audio(data)
                    for frame in encrypted_frames:
                        # Decrypt and decode
                        audio_samples = self.voice_protocol.decrypt_and_decode(frame)
                        if audio_samples:
                            self.audio_received.emit(audio_samples, self.client_id)
        except Exception as e:
            print(f"Client {self.client_id} GSM data processing error: {e}")
    
    def _handle_encrypted_message(self, plaintext):
        """Handle decrypted message content."""
        # Check if it's audio data or control message
        if plaintext.startswith(b'AUDIO:'):
            audio_data = plaintext[6:]
            self.data_received.emit(audio_data, self.client_id)
        else:
            # Control message
            try:
                message = plaintext.decode('utf-8')
                self._handle_control_message(message)
            except:
                # Binary data
                self.data_received.emit(plaintext, self.client_id)
    
    def _handle_control_message(self, message):
        """Handle control messages."""
        if message == "RINGING":
            self.state_changed.emit("RINGING", "", self.client_id)
        elif message == "IN_CALL":
            self.state_changed.emit("IN_CALL", "", self.client_id)
        elif message == "CALL_END":
            self.state_changed.emit("CALL_END", "", self.client_id)
    
    def _handle_voice_start(self, voice_start_msg):
        """Handle voice session start."""
        if voice_start_msg:
            # Initialize voice protocol with negotiated parameters
            self.protocol.voice_session_active = True
            self.protocol.voice_session_id = voice_start_msg.session_id
            
            # Send acknowledgment
            self.protocol.send_voice_ack(voice_start_msg.session_id)
            
            # Initialize voice codec
            self._initialize_voice_protocol(voice_start_msg.codec_mode)
            self.voice_active = True
    
    def _initialize_voice_protocol(self, codec_mode=Codec2Mode.MODE_1200):
        """Initialize voice protocol with codec and encryption."""
        if self.protocol.hkdf_key:
            self.voice_protocol = VoiceProtocol(
                shared_key=bytes.fromhex(self.protocol.hkdf_key),
                codec_mode=codec_mode,
                cipher_type=self.protocol.cipher_type
            )
            print(f"Client {self.client_id} initialized voice protocol")
    
    def initiate_call(self):
        """Initiate a call to the peer."""
        if not self.peer_port:
            print(f"Client {self.client_id}: No peer port set")
            return False
            
        # Connect to peer
        self.protocol.connect_to_peer(self.peer_port)
        self.state_changed.emit("CALLING", "", self.client_id)
        
        # Start key exchange
        self.protocol.generate_ephemeral_keys()
        self.protocol.send_ping_request(cipher_type=1)  # Request ChaCha20
        
        return True
    
    def answer_call(self):
        """Answer an incoming call."""
        self.state_changed.emit("IN_CALL", "", self.client_id)
        
        # Enable auto-responder for handling protocol flow
        self.protocol.auto_responder = True
        
        # If we already have a ping request, respond to it
        for i, msg in enumerate(self.protocol.inbound_messages):
            if msg.get('type') == 'PING_REQUEST' and not self.protocol.state['ping_sent']:
                self.protocol.respond_to_ping(i, 1)  # Accept with ChaCha20
                break
    
    def end_call(self):
        """End the current call."""
        if self.voice_active:
            self.protocol.end_voice_call()
        
        self.voice_active = False
        self.handshake_complete = False
        self.state_changed.emit("CALL_END", "", self.client_id)
        
        # Close connections
        for conn in self.protocol.connections:
            conn.close()
        self.protocol.connections.clear()
    
    def send_audio(self, audio_samples):
        """Send audio samples through the voice protocol."""
        if self.voice_active and self.voice_protocol and self.gsm_socket:
            # Encode and encrypt audio
            fsk_audio = self.voice_protocol.encode_and_encrypt(audio_samples)
            if fsk_audio and self.gsm_socket:
                try:
                    # Send FSK-modulated audio through GSM simulator
                    self.gsm_socket.send(fsk_audio)
                except Exception as e:
                    print(f"Client {self.client_id} failed to send audio: {e}")
    
    def send_message(self, message):
        """Send an encrypted message."""
        if self.handshake_complete:
            if isinstance(message, str):
                message = message.encode('utf-8')
            self.protocol.send_encrypted_message(message)
    
    def start_voice_session(self):
        """Start a voice session."""
        if self.handshake_complete and not self.voice_active:
            self._initialize_voice_protocol()
            self.protocol.start_voice_call(codec_mode=5)  # 1200 bps mode
    
    def get_identity_key(self):
        """Get this client's identity public key."""
        if hasattr(self.protocol, 'identity_pubkey') and self.protocol.identity_pubkey:
            return self.protocol.identity_pubkey.hex()
        return "test_identity_key_" + str(self.client_id)
    
    def get_local_port(self):
        """Get the local listening port."""
        if hasattr(self.protocol, 'local_port'):
            return self.protocol.local_port
        return 12345 + self.client_id
    
    def stop(self):
        """Stop the client."""
        self.running = False
        
        # End voice session if active
        if self.voice_active:
            self.protocol.end_voice_call()
        
        # Stop protocol server listener
        if hasattr(self.protocol, 'server_listener') and self.protocol.server_listener:
            self.protocol.server_listener.stop()
        
        # Close GSM socket
        if self.gsm_socket:
            try:
                self.gsm_socket.close()
            except:
                pass
                
        self.quit()
        self.wait(1000)